from qiskit import IBMQ
IBMQ.load_account()
from kaleidoscope.qiskit.backends.interactive import system_gate_map

pro = IBMQ.get_provider(group='open')
backend = pro.backends.ibmq_vigo
system_gate_map(backend, as_widget=True)

system_gate_map(backend, qubit_labels=['A', 'B', 'C', 'D', 'E'],
                as_widget=True)

system_gate_map(backend,
             qubit_labels=['A', 'B', 'C', 'D', 'E'],
             qubit_colors='#8b7b8b',
             as_widget=True)

system_gate_map(backend,
             qubit_labels=['A', 'B', 'C', 'D', 'E'],
             qubit_colors=['#8b7b8b', '#845e49', '#496b3a', '#e97fa5', '#ff9999'],
             as_widget=True)

system_gate_map(backend,
             qubit_colors='#8b7b8b',
             line_colors='#ff9999',
             as_widget=True)

system_gate_map(backend,
             line_colors=['#845e49', '#496b3a', '#e97fa5', '#ff9999']*2,
             as_widget=True)

system_gate_map(backend, label_qubits=False, as_widget=True)

system_gate_map(backend,
                qubit_colors='white',
                font_color="black",
                line_colors='white',
                background_color='black',
                as_widget=True)


import numpy as np
from matplotlib.colors import LinearSegmentedColormap, rgb2hex
from kaleidoscope.interactive import bloch_sphere

cm = LinearSegmentedColormap.from_list('graypurple', ["#999999", "#AA00FF"])

pointsx = [[0, -np.sin(th), np.cos(th)] for th in np.linspace(0, np.pi/2, 20)]
pointsz = [[np.sin(th), -np.cos(th), 0] for th in np.linspace(0, 3*np.pi/4, 30)]

points = pointsx + pointsz

points_alpha = [np.linspace(0.8, 1, len(points))]

points_color = [[rgb2hex(cm(kk)) for kk in np.linspace(-1,1,len(points))]]

vectors_color = ["#777777", "#AA00FF"]

bloch_sphere(points=points,
             vectors=[[0, 0, 1], [1/np.sqrt(2), 1/np.sqrt(2), 0]],
             vectors_color=vectors_color,
             points_alpha=points_alpha,
             points_color=points_color,
             as_widget=True)


from kaleidoscope.qiskit.overload import QuantumCircuit
from kaleidoscope.qiskit.providers import simulators
from kaleidoscope.interactive import counts_distribution

sim = simulators.aer_vigo_simulator

qc = QuantumCircuit(3, 3) >> sim
qc.h(1)
qc.cx(1,0)
qc.cx(1,2)
qc.measure(range(3), range(3))

counts = qc.transpile().sample().result_when_done()
counts_distribution(counts, as_widget=True)


from qiskit import IBMQ
IBMQ.load_account()
from kaleidoscope.qiskit.backends.interactive import system_error_map

pro = IBMQ.get_provider(group='open')
backend = pro.backends.ibmq_vigo
system_error_map(backend, as_widget=True)

import datetime
# Grab ibmq_vigo properties on Jan. 1, 2020.
old_props = backend.properties(datetime=datetime.datetime(2020, 1, 1))
system_error_map(old_props, as_widget=True)


import numpy as np
from qiskit import *
from qiskit.quantum_info import Statevector
from kaleidoscope.interactive import bloch_multi_disc

N = 4
qc = QuantumCircuit(N)
qc.h(range(N))
for kk in range(N):
    qc.ry(2*np.pi*np.random.random(), kk)
for kk in range(N-1):
    qc.cx(kk,kk+1)
for kk in range(N):
    qc.rz(2*np.pi*np.random.random(), kk)

state = Statevector.from_instruction(qc)
bloch_multi_disc(state, as_widget=True)


import numpy as np
from kaleidoscope import bloch_sphere

vec = [0, 1/np.sqrt(2) , 1/np.sqrt(2)]
bloch_sphere(vec, as_widget=True)

vec = [1/np.sqrt(2), 0, 1/np.sqrt(2)]
bloch_sphere(vec, vectors_color='#ff0000', vectors_alpha=0.25,
             as_widget=True)

vec1 = [0, 0, 1]
vec2 = [1, 0, 0]
vec3 = [1/np.sqrt(2), 1/np.sqrt(2), 0]
vec4 = [0, 1/np.sqrt(2), 1/np.sqrt(2)]

bloch_sphere([vec1, vec2, vec3, vec4], as_widget=True)

vec1 = [0, 0, 1]
vec2 = [1, 0, 0]
vec3 = [1/np.sqrt(2), 1/np.sqrt(2), 0]
vec4 = [0, 1/np.sqrt(2), 1/np.sqrt(2)]

bloch_sphere([vec1, vec2, vec3, vec4],
             vectors_color=['#e34234', '#6f4e37', '#00008b', '#ff1493'],
             vectors_alpha=[1.0,0.35, 0.1, 0.95],
             as_widget=True)

vec1 = [0, 0, 1]
vec2 = [1, 0, 0]
vec3 = [1/np.sqrt(2), 1/np.sqrt(2), 0]
vec4 = [0, 1/np.sqrt(2), 1/np.sqrt(2)]

bloch_sphere([vec1, vec2, vec3, vec4], vectors_annotation=True,
             as_widget=True)

vec1 = [0, 0, 1]
vec2 = [1, 0, 0]
vec3 = [1/np.sqrt(2), 1/np.sqrt(2), 0]
vec4 = [0, 1/np.sqrt(2), 1/np.sqrt(2)]

bloch_sphere([vec1, vec2, vec3, vec4],
             vectors_annotation=[False, True, False, True],
             as_widget=True)

vec = [1/np.sqrt(2), 0, 1/np.sqrt(2)]
bloch_sphere(points=vec, as_widget=True)

vec = [1/np.sqrt(2), 0, 1/np.sqrt(2)]
bloch_sphere(points=vec, points_color='#ff0000',
             as_widget=True)

vec1 = [0, 0, 1]
vec2 = [1, 0, 0]
vec3 = [1/np.sqrt(2), 1/np.sqrt(2), 0]
vec4 = [0, 1/np.sqrt(2), 1/np.sqrt(2)]

bloch_sphere(points=[vec1, vec2, vec3, vec4], as_widget=True)

vec1 = [0, 0, 1]
vec2 = [1, 0, 0]
vec3 = [1/np.sqrt(2), 1/np.sqrt(2), 0]
vec4 = [0, 1/np.sqrt(2), 1/np.sqrt(2)]

bloch_sphere(points=[vec1, vec2, vec3, vec4],
            points_color=['#e34234', '#6f4e37', '#00008b', '#8014ff'],
            as_widget=True)

vec1 = [0, 0, 1]
vec2 = [1, 0, 0]
vec3 = [1/np.sqrt(2), 1/np.sqrt(2), 0]
vec4 = [0, 1/np.sqrt(2), 1/np.sqrt(2)]

bloch_sphere(points=[[vec1, vec2, vec3, vec4]],
            points_color=[['#e34234', '#6f4e37', '#00008b', '#8014ff']],
            as_widget=True)

vec1 = [0, 0, 1]
vec2 = [1, 0, 0]
vec3 = [1/np.sqrt(2), 1/np.sqrt(2), 0]
vec4 = [0, 1/np.sqrt(2), 1/np.sqrt(2)]

bloch_sphere(points=[[vec1, vec2], [vec3, vec4]], as_widget=True)

vec1 = [0, 0, 1]
vec2 = [1, 0, 0]
vec3 = [1/np.sqrt(2), 1/np.sqrt(2), 0]
vec4 = [0, 1/np.sqrt(2), 1/np.sqrt(2)]

bloch_sphere(points=[[vec1, vec2], [vec3, vec4]],
            points_color=['#e34234', '#8014ff'],
            as_widget=True)

vec1 = [0, 0, 1]
vec2 = [1, 0, 0]
vec3 = [1/np.sqrt(2), 1/np.sqrt(2), 0]
vec4 = [0, 1/np.sqrt(2), 1/np.sqrt(2)]

bloch_sphere(points=[[vec1, vec2], [vec3, vec4]],
            points_color=[['#e34234', '#ff8014'], ['#8014ff', '#93ff14']],
            as_widget=True)

vec1 = [0, 0, 1]
vec2 = [1, 0, 0]
vec3 = [1/np.sqrt(2), 1/np.sqrt(2), 0]
vec4 = [0, 1/np.sqrt(2), 1/np.sqrt(2)]

bloch_sphere(points=[vec1, vec2, vec3, vec4],
            points_color=['#e34234', '#6f4e37', '#00008b', '#8014ff'],
            points_alpha=[1.0, 0.5, 1.0, 0.6], as_widget=True)

from matplotlib.colors import LinearSegmentedColormap, rgb2hex
cm = LinearSegmentedColormap.from_list('graypurple', ["#999999", "#AA00FF"])

pointsx = [[0, -np.sin(th), np.cos(th)] for th in np.linspace(0,np.pi/2,20)]
pointsz = [[np.sin(th), -np.cos(th), 0] for th in np.linspace(0,3*np.pi/4,30)]
points = pointsx+pointsz

points_alpha = [np.linspace(0.8,1, len(points))]
points_color = [[rgb2hex(cm(kk)) for kk in np.linspace(-1,1,len(points))]]
vectors_color = ["#777777", "#AA00FF"]

bloch_sphere(points=points,
            vectors=[[0,0,1], [1/np.sqrt(2),1/np.sqrt(2),0]],
            vectors_color=vectors_color,
            points_alpha=points_alpha,
            points_color=points_color,
            as_widget=True)


import numpy as np
from qiskit import *
from qiskit.quantum_info import Statevector

from kaleidoscope.interactive import bloch_disc
qc = QuantumCircuit(1)
qc.ry(np.pi*np.random.random(), 0)
qc.rz(np.pi*np.random.random(), 0)

state = Statevector.from_instruction(qc)
bloch_disc(state, as_widget=True)


# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Overloading of the QuantumCircuit class in Qiskit"""

import copy
from types import MethodType
import qiskit
from qiskit.compiler.assemble import assemble
from qiskit.compiler import transpile
from qiskit.quantum_info import Statevector, Operator
from qiskit.providers.ibmq.ibmqbackend import IBMQBackend
from qiskit.providers.basebackend import BaseBackend
from qiskit.tools.monitor import job_monitor as aer_monitor
from qiskit.providers.ibmq.job import job_monitor as ibmq_monitor


class QuantumCircuit(qiskit.circuit.quantumcircuit.QuantumCircuit):
    """A QuantumCircuit class that contains additional functionality.

    Attributes:
        target_backend (BaseBackend): A backend to target.
        is_kal (bool): Is this an overloaded Kaleidoscope circuit.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.target_backend = None
        self.is_kal = True

    def __rshift__(self, target):
        """Add a target backend to circuit.

        Parameters:
            target (BaseBackend): The target backend

        Returns:
            QuantumCircuit: QuantumCircuit with attached target_backend.

        Raises:
            TypeError: Input is not a valid backend instance.
        """
        if not isinstance(target, BaseBackend):
            raise TypeError('Target is not a valid backend instance.')
        self.target_backend = target  # pylint: disable=attribute-defined-outside-init
        return self

    def _wrap_circuit(self, circ, inplace=False):
        """Make an overloaded circuit from a standard Qiskit one.

        Parameters:
            circ (QuantumCircuit): Input QuantumCircuit.
            inplace (bool): Return a copy of the circuit data.

        Returns:
            QuantumCircuit: Returns overloaded QuantumCircuit.

        Raises:
            TypeError: Input is not a QuantumCircuit instance.
        """
        if not isinstance(circ, qiskit.circuit.quantumcircuit.QuantumCircuit):
            raise TypeError('Input should be a QuantumCircuit')
        out_qc = QuantumCircuit()
        if inplace:
            out_qc.name = copy.copy(circ.name)
            out_qc.qregs = copy.copy(circ.qregs)
            out_qc.cregs = copy.copy(circ.cregs)
            out_qc._data = copy.deepcopy(circ._data)
            out_qc._parameter_table = copy.deepcopy(circ._parameter_table)
            out_qc._layout = copy.copy(circ._layout)
        else:
            out_qc.name = circ.name
            out_qc.qregs = circ.qregs
            out_qc.cregs = circ.cregs
            out_qc._data = circ._data
            out_qc._parameter_table = circ._parameter_table
            out_qc._layout = circ._layout
        return out_qc

    def sample(self, backend=None, shots=1024, seed_simulator=None, memory=False):
        """Sample a the output distribution from a quantum circuit.

        Parameters:
            backend (BaseBackend): Backend to use.  Default is Aer QASM simulator.
            shots (imt): Number of times to sample.  Default is 1024.
            seed_simulator (int):  Seed to use for simulator (if backend is a simulator).
            memory (bool): Return individual measurement results.

        Returns:
            job: A job instance with `block_until_ready` attribute.
        """
        if backend is None:
            if self.target_backend:
                backend = self.target_backend
            else:
                backend = qiskit.Aer.get_backend('qasm_simulator')

        qobj = assemble(self,
                        shots=shots,
                        memory=memory,
                        seed_simulator=seed_simulator,
                        backend=backend)

        job = backend.run(qobj)

        if isinstance(backend, IBMQBackend):
            job.result_when_done = MethodType(ibmq_wait, job)
        else:
            job.result_when_done = MethodType(aer_wait, job)
        return job

    def statevector(self, include_final_measurements=False):
        """Return output statevector for the circuit.

        Parameters:
            include_final_measurements (bool): Include final measurements
                                               in the circuit.

        Returns:
            StateVector: Output statevector object.
        """
        new_circ = self
        if not include_final_measurements:
            new_circ = self.remove_final_measurements(inplace=False)
        return Statevector.from_instruction(new_circ)

    def unitary(self):
        """Return the unitary for the circuit (if any).

        Returns:
            Operator: Unitary for the circuit.
        """
        new_circ = self.remove_final_measurements(inplace=False)
        return Operator(new_circ)

    def transpile(self, backend=None, **kwargs):
        """Transpile the circuit.

        If the circuit has a `target_backend` assigned
        and `backend=None` then the target backend is
        used for transpilation.  All other `transpile`
        functionality remains unchanged.

        Parameters:
            backend (BaseBackend): Backend to transpile for.
            kwargs: any other transpiler kwargs.

        Returns:
            QuantumCircuit: Transpiled quantum circuit.

        Returns:

        """
        if backend is None:
            if self.target_backend:
                backend = self.target_backend
        new_qc = transpile(self, backend=backend, **kwargs)
        out = self._wrap_circuit(new_qc, inplace=True)
        out.target_backend = self.target_backend
        return out


def aer_wait(self, monitor=False):
    """Monitor air jobs

    Parameters:
        monitor (bool): Monitor the job.

    Returns:
        dict: Counts data
    """
    if monitor:
        aer_monitor(self, interval=1e-3)
    return self.result().get_counts()


def ibmq_wait(self, monitor=False):
    """Monitor IBMQ jobs

    Parameters:
        monitor (bool): Monitor the job.

    Returns:
        dict: Counts data
    """
    if monitor:
        ibmq_monitor(self)
    return self.result().get_counts()



from qiskit import *
from kaleidoscope.qiskit.backends import cnot_error_density
provider = IBMQ.load_account()

backends = provider.backends(simulator=False,
                             filters=lambda b: b.configuration().n_qubits == 5)

cnot_error_density(backends)


# -*- coding: utf-8 -*-

# This code is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""General plotting utility functions"""


def hex_to_rgb(hex_color):
    """Converts a HEX color to a tuple of RGB values.

    Parameters:
        hex_color (str): Input hex color string.

    Returns:
        tuple: RGB color values.
    """
    hex_color = hex_color.lstrip("#")
    if len(hex_color) == 3:
        hex_color = hex_color * 2
    return int(hex_color[0:2], 16), int(hex_color[2:4], 16), int(hex_color[4:6], 16)
